/* $Id: virtual_stub.c,v 1.21 2003/08/30 22:31:39 mikpe Exp $
 * Kernel stub used to support virtual perfctrs when the
 * perfctr driver is built as a module.
 *
 * Copyright (C) 2000-2003  Mikael Pettersson
 */
#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/fs.h>
#include <linux/perfctr.h>
#include <linux/kmod.h>
#include "compat.h"

static void bug_void_perfctr(struct vperfctr *perfctr)
{
	task_thread(current)->perfctr = NULL;
	BUG();
}

static int vperfctr_stub_open(struct inode *inode, struct file *filp)
{
	struct file_operations *fops;

	read_lock(&vperfctr_stub_lock);
	fops = fops_get(vperfctr_stub.file_ops);
	read_unlock(&vperfctr_stub_lock);
	if( !fops && request_module("perfctr") == 0 ) {
		read_lock(&vperfctr_stub_lock);
		fops = fops_get(vperfctr_stub.file_ops);
		read_unlock(&vperfctr_stub_lock);
	}
	if( !fops )
		return -ENOSYS;
	filp->f_op = fops;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
	inode->i_fop = fops; /* no fops_get since only filp->f_op counts */
#endif
	return fops->open(inode, filp);
}

static struct file_operations vperfctr_stub_file_ops = {
	.open = vperfctr_stub_open,
};

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
void perfctr_set_proc_pid_ops(struct inode *inode)
{
	inode->i_fop = &vperfctr_stub_file_ops;
}
#else
#include <linux/proc_fs.h>
struct inode_operations perfctr_proc_pid_inode_operations = {
	.default_file_ops = &vperfctr_stub_file_ops,
	.permission = proc_permission,
};
#endif

struct vperfctr_stub vperfctr_stub = {
	.exit = bug_void_perfctr,
	.suspend = bug_void_perfctr,
	.resume = bug_void_perfctr,
	.sample = bug_void_perfctr,
	.file_ops = NULL,
};
rwlock_t vperfctr_stub_lock = RW_LOCK_UNLOCKED;

/*
 * exit_thread() calls __vperfctr_exit() via vperfctr_stub.exit().
 * If the process' reference was the last reference to this
 * vperfctr object, and this was the last live vperfctr object,
 * then the perfctr module's use count will drop to zero.
 * This is Ok, except for the fact that code is still running
 * in the module (pending returns back to exit_thread()). This
 * could race with rmmod in a preemptive UP kernel, leading to
 * code running in freed memory. The race also exists in SMP
 * kernels, but the time window is extremely small.
 *
 * Since exit() isn't performance-critical, we wrap the call to
 * vperfctr_stub.exit() with code to increment the module's use
 * count before the call, and decrement it again afterwards. Thus,
 * the final drop to zero occurs here and not in the module itself.
 * (All other code paths that drop the use count do so via a file
 * object, and VFS in 2.4+ kernels also refcount the module.)
 */
void _vperfctr_exit(struct vperfctr *perfctr)
{
	__module_get(vperfctr_stub.owner);
	vperfctr_stub.exit(perfctr);
	module_put(vperfctr_stub.owner);
}

EXPORT_SYMBOL(vperfctr_stub);
EXPORT_SYMBOL(vperfctr_stub_lock);
EXPORT_SYMBOL_pidhash;
EXPORT_SYMBOL___put_task_struct;
#ifdef CONFIG_SMP
EXPORT_SYMBOL_tasklist_lock;
#endif
